$TYPECHECK ON
'------------------------------------------------------------------------------------------------------
'
' purpose of this program is to get sample matching measurements of retinal nerve fiber layer (RNFL)
' or remeasure misaligned scans to get estimates from a scaled circle scan
'  impetus for this comes from the article by 
'       CARPINETO, CIANCAGLINI, AHARRH-GNAMA, CIRONE, and MASTROPA
'       Custom measurement of retinal nerve fiber layer thickness using STRATUS OCT in normal eyes
'       European Journal of Ophthalmology / Vol. 15 no. 3, 2005 / pp. 360-366
' showing that smaller optic nerves have smaller global RNFL by artifact of using a 
' set circular scan size. If you could make the circular scan smaller for smaller disks 
' then there is no relationship of global RNFL to optic disk size
' the authors of that article made the radius of the custom scans 850 microns from the disk edge
' therefore, if a disk was 1400 microns wide, the new scan radius would be
'  700 + 850 = 1550 microns
'
' First use of the program was published in 
'        Kelly JP, Baran F, Phillips JO, Weiss AH. 
'        Optical Coherence Tomography in Optic Nerve Hypoplasia: 
'        Correlation With Optic Disc Diameter, Nerve Fiber Layer Thickness, and Visual Function. 
'        J Neuroophthalmol. 2018;38(3):312-319. 
'
'
'  Version 1.1
'  July 2016, John Kelly
'   Added auto segmentation of ILM and NFL
'   Fixed "refresh" to update measurments from new ILM/NFL modifications
'   Allow toggle On/Off of selecting each ILM/NFL layers
'   Add psuedo OnDblClick event to add a control point, like Heyex
'
'  Version 1.2
'  January 2017, John Kelly
'   Added 'Save Image' for data plotted on normative template
'
'  Version 1.3
'  October 2017, John Kelly
'   Wrote filtering routines on BitampGfx
'   Median filter uses 8-bit grey scale for speed
'   Fixed exit bug
'
'  Version 1.4.0
'  Febrary 2019, John Kelly
'   Uses OpenCV wrapper for faster pixel processing
'   Added option to set a reference scan and check disc movement from reference
'
'  Version 1.4.1 revision
'  June 2019, John Kelly
'   reference, savedata, and result View added an error point at the end
'   this was due to a bug for zero based arrays compared to counter, e.g.
'       FOR i = 0 TO OCTLocus(0).Num
'   should have been:
'       FOR i = 0 TO OCTLocus(0).Num -1 
'
'  Version 1.4.2 revision
'  June 2020, John Kelly
'   fixed minor bug of missing initial right-mouse double-click on OCT image
'
'
'  Version 1.4.3 revision
'   July 2020, John Kelly
'   removed dependences of BitmapGfx.inc. Now all background bitmap
'   operations are performed by OpenCV
'
'------------------------------------------------------------------------------------------------------


$TYPECHECK ON

$INCLUDE <Rapidq2.inc>
$INCLUDE <fileops.inc>
DIM d as QDEBUG

$INCLUDE <Manual_RNFL_Decl.inc>
$INCLUDE <OpenCV_wrapper.inc>
$INCLUDE <Manual_RNFL_Globals.inc>
$INCLUDE <Manual_RNFL_Form.inc>
$INCLUDE <Manual_RNFL_Norm.inc>
$INCLUDE <Manual_RNFL_Reference.inc>
$INCLUDE <Manual_RNFL_Spline.inc>
$INCLUDE <Manual_RNFL_Segmentation.inc>
$INCLUDE <Polynomial_Fit.inc>
$INCLUDE <DiscTracking.inc>
$INCLUDE <Manual_RNFL_XPButtons.inc>




'-------------------------------------------------
' proper end of app, save data and release objects
'-------------------------------------------------
FUNCTION CheckEndProg() AS INTEGER
    RESULT = False
    DEFSTR TheMsg = "Exit application?" + CRLF
    IF OutputFileName <> "" THEN 
        TheMsg = TheMsg + "Your data will be stored in " + OutputFileName
    ELSE
        IF OCTLocus(0).Num > 0 THEN TheMsg = TheMsg + "You will be asked to save current data"
    END IF

    IF MessageBox(TheMsg, "Exit...", &H30 + 4) = mrYes THEN 
        IF OutputFileName <> "" THEN 
            IF MessageBox("Save data of " + Fname + " " + LName + CRLF + _
            "plotted on normative template?", "Save image...", &H30 + 4) = mrYes THEN NormForm_Save
        END IF
        IF OCTLocus(0).Num > 0 THEN StoreData
        RESULT = True
    END IF
END FUNCTION




SUB Form_Close(ByRef Action AS INTEGER)
    IF CheckEndProg = True THEN
        Application.Terminate
    ELSE
        Action = False
    END IF
END SUB



SUB EndButn_Click
    Form.Close      'proper event handling
END SUB




'-------------------------------------------------------
'pair the loci on scan to the spline fit
'-------------------------------------------------------
SUB ParseSegmentation
    DEFINT i, j, Lay, xDiff, x1, x2, FoundMatch
    DEFINT myLayer = CurrLayer                                              'temp store current layer
    DIM yp(MaxLayers) AS SINGLE

    FOR i = 0 TO OCTLocus(0).Num -1                                         'for all matched loci, 0 based index
        OCTLocus(i).RNFL_Thickness = 0                                      'set to no match yet
        FoundMatch = -1                                                     'count matches for each segmentation, 0 based index
        FOR Lay = 0 TO MaxLayers                                            'for each segmentation
            CurrLayer = Lay                                                 'keep variable selection consistent for other subs
            yp(Lay) = 0                                                     'clear each measurement until a match found
            IF CptX(Lay, 0) > CptX(Lay, CptN(Lay)) THEN ReverseSpline       'keep a consistent control point Left -> Right order
            FOR j = 0 TO SegN(Lay)-1                                        'ILM segmentation line
                x1 = SegX(Lay, j) - OCT_Left
                x2 = SegX(Lay, j+1) - OCT_Left                              'correct for displacement
                IF x1 = OCTLocus(i).ScreenX THEN
                    yp(Lay) = SegY(Lay, j)
                    FoundMatch ++
                    EXIT FOR
                ELSE
                    'spline data should already be sorted into low to high ordering (compensated for right-left drawing)
                    IF (x1 < OCTLocus(i).ScreenX) AND (x2 >= OCTLocus(i).ScreenX) THEN
                        xDiff = x2-x1
                        IF xDiff <= 0 THEN xDiff = 1
                        'linear interpolation between points
                        yp(Lay) = SegY(Lay, j) + ((SegY(Lay, j+1) - SegY(Lay, j)) * (OCTLocus(i).ScreenX - x1) / xDiff)
                        FoundMatch ++
                        EXIT FOR
                    END IF
                END IF
            NEXT j
        NEXT Lay

        IF FoundMatch = MaxLayers THEN
            OCTLocus(i).RNFL_Thickness  = IIF(LayerShowBox(cNFL).Checked, ABS(yp(cNFL) - yp(cILM)) * 1000.0! * OCTmmPerYPixels / OCTZoomFactor, 0)  'store thickness from start-end click in microns
            OCTLocus(i).ONL_Thickness   = IIF(LayerShowBox(cONL).Checked, ABS(yp(cONL) - yp(cILM)) * 1000.0! * OCTmmPerYPixels / OCTZoomFactor, 0)  'store thickness from start-end click in microns
            OCTLocus(i).Total_Thickness = IIF(LayerShowBox(cBM).Checked,  ABS(yp(cBM)  - yp(cILM)) * 1000.0! * OCTmmPerYPixels / OCTZoomFactor, 0)  'store thickness from start-end click in microns
        ELSE
            OCTLocus(i).RNFL_Thickness  = 0
            OCTLocus(i).ONL_Thickness   = 0
            OCTLocus(i).Total_Thickness = 0
        END IF
    NEXT i

    CurrLayer = myLayer     'restore
END SUB




SUB StoreData
    DEFSTR c = ","
    DEFSTR sTmp = ""
    DataList.Clear

    MySaveDialog.Caption = "Save Data?"
    IF OutputFileName = "" THEN
        MySaveDialog.InitialDir = CURDIR$
        MySaveDialog.Filter = "Output files|*.txt;*.csv|All Files|*.*"
        MySaveDialog.FilterIndex = 0
        MySaveDialog.FileName = FName + "_" + LName + "_" + Eye + "_" + MRN
        IF MySaveDialog.Execute THEN 
            OutputFileName = MySaveDialog.FileName
        ELSE
            EXIT SUB
        END IF
    END IF

    IF OutputFileName = "" THEN EXIT SUB                            'user bails out

    DEFSTR myExt = StripFileExt(OutputFileName)
    IF myExt = "" THEN OutputFileName = OutputFileName + ".csv"     'add extension if none

   'create a header
    IF FILEEXISTS(OutputFileName) = False THEN
        sTmp = _
        "LastName,FirstName," + _
        "MRN," + _
        "Eye," + _
        "OriginalRadius_um,NewScanRadius_um," + _
        "RNFL_Angle" + degr + "," + _
        "RNFL_Thickness_um," + _
        "ONL_Thickness_um," + _
        "Total_Thickness_um," + _
        "DiscX_mm,DiscY_mm," + _
        "Xpix_on_SLO,Ypix_on_SLO," + _
        "Fovea_X,Fovea_Y," + _
        "Fovea_ang" + degr + "," + _
        "Focus,Quality,ARTaverage," + _
        "Met_PositionTolerance?"
        IF MyImageMatch.centerX <> 0 THEN
            sTmp = sTmp + ",DiscMatchRotation" + degr + "," + _
                "DiscMatchScaleX,DiscMatchScaleY," + _
                "DiscMatchCenterX,DiscMatchCenterY," + _
                "DiscMatchShear,DiscMatchNumMatches"
        END IF
        sTmp = sTmp + CRLF 
        DataList.Text = sTmp
    ELSE
        DataList.LoadFromFile(OutputFileName)
        sTmp = DataList.Text                                                        'append if it exists
    END IF

    IF ScanType = "CIRCLE" AND (SplineButton.Caption = "Spline") THEN 
        ParseSegmentation           'convert segmentation lines to RNFL thickness, this will clear all 'lines' segment measurements!!
    END IF

    DEFINT i
    FOR i = 0 TO OCTLocus(0).Num -1                                                 '0 based arrays
        IF OCTLocus(i).RNFL_Thickness > 0 THEN
            sTmp = sTmp + LName + c + FName + c + MRN + c + Eye + c                 'name, MRN, and eye
            IF ScanType = "CIRCLE" THEN                                             'origninal scan radius
                sTmp = sTmp + STRF$(Circ.Radius * 1000.0!, ffFixed, 10, 1)  + c
            ELSE
                sTmp = sTmp + "Section" + c
            END IF
            sTmp = sTmp + ScanRadiusEdit.Text + c                                   'new scan radius
            sTmp = sTmp + STRF$( OCTLocus(i).Angle, ffFixed, 10, 3) + c             'angle of thickness measurment
            sTmp = sTmp + STRF$( OCTLocus(i).RNFL_Thickness, ffFixed, 10, 1) + c    'RNFL thickness at that angle
            sTmp = sTmp + STRF$( OCTLocus(i).ONL_Thickness, ffFixed, 10, 1) + c     'RNFL thickness at that angle
            sTmp = sTmp + STRF$( OCTLocus(i).Total_Thickness, ffFixed, 10, 1) + c   'RNFL thickness at that angle
            sTmp = sTmp + STRF$( DiskCenter.X, ffFixed, 10, 1) + c                  'location of disk (mm) on SLO
            sTmp = sTmp + STRF$( DiskCenter.Y, ffFixed, 10, 1) + c                  'in X, Y coordinates 0,0 at top left
            sTmp = sTmp + STRF$( OCTLocus(i).SLOscreenX , ffFixed, 10, 1) + c       'location of measurement on SLO in pixels
            sTmp = sTmp + STRF$( OCTLocus(i).SLOScreenY, ffFixed, 10, 1) + c        'in X, Y coordinates 0,0 at top left
            sTmp = sTmp + STRF$( FovCenter.X, ffFixed, 10, 1) + c                   'location of fovea on SLO in pixels
            sTmp = sTmp + STRF$( FovCenter.Y, ffFixed, 10, 1) + c                   'in X, Y coordinates 0,0 at top left
            sTmp = sTmp + STRF$(FoveaAngle(), ffFixed, 10, 1) + c                   'angle of fovea from disk center (inferior to disk = negative angle)
            sTmp = sTmp + SLOFocus + c                                              'focus of SLO
            sTmp = sTmp + OCTQuality + c                                            'OCT image quality score 
            sTmp = sTmp + ARTaverage + c                                            'average number of images making up OCT with automated real-time tracking
            sTmp = sTmp + OCTPositionTolerance                                      'OCT acquired within position tolerance?
            IF MyImageMatch.centerX <> 0 THEN
                sTmp = sTmp + c
                sTmp = sTmp + STR$(MyImageMatch.rotationDeg) + c
                sTmp = sTmp + STR$(MyImageMatch.scaleX) + "," + STR$(MyImageMatch.scaleY) + c
                sTmp = sTmp + STR$(MyImageMatch.CenterX) + "," + STR$(MyImageMatch.CenterY) + c
                sTmp = sTmp + STR$(MyImageMatch.shear) + c
                sTmp = sTmp + STR$(MyImageMatch.NumMatchPoints) 
            END IF
            sTmp = sTmp + CRLF
        END IF
    NEXT i
    'sTmp = sTmp + CRLF

    DataList.Text = sTmp
    DataList.SaveToFile(OutputFileName)
    DOEVENTS
    NormForm_Paint
END SUB




SUB SaveButtn_Click
    IF ReferenceCanBeSaved THEN 
        StoreReferenceData
    ELSE
        StoreData
    END IF

    SkipButtn_Click         'signal click to load next for main loop
END SUB



SUB SkipButtn_Click
    ClearAllData                'clear out for next data set, don't store data
    SaveButn.Tag = True         'signal click to load next for main loop
END SUB



SUB DeleteAllSplines
    'clear segmentation splines
    MEMSET( VARPTR( SegX(0, 0)),       0, (MaxLayers+1) * (MaxPts+1) * SIZEOF(INTEGER))
    MEMSET( VARPTR( SegY(0, 0)),       0, (MaxLayers+1) * (MaxPts+1) * SIZEOF(INTEGER))
    MEMSET( VARPTR( SegN(0)),          0, (MaxLayers+1)              * SIZEOF(INTEGER))
    'clear control points
    MEMSET( VARPTR( CptX(0, 0)),       0, (MaxLayers+1) * (MaxCtrlPts+1) * SIZEOF(INTEGER))
    MEMSET( VARPTR( CptY(0, 0)),       0, (MaxLayers+1) * (MaxCtrlPts+1) * SIZEOF(INTEGER))
    MEMSET( VARPTR( CptN(0)),         -1, (MaxLayers+1)                  * SIZEOF(INTEGER))  '-1 signals none
    MEMSET( VARPTR( SplineLocked(0)),  0, (MaxLayers+1)                  * SIZEOF(INTEGER)) 

    ShowSplineControlPoints(False)
    CptN(CurrLayer) = cNone             'set to none
    SplineLocked(CurrLayer) = False
    CtrlPtDrag = False
    DrawFlag = False
END SUB



SUB ClearAllData
    ClearLoci
    DeleteAllSplines
    SelectedIndx = -1
    FoveaHasBeenSet = False
    DiscHasBeenSet = False
    DeleteKeyDown = False
    LastCtrlPt = -1
    TweakOCT.Draw(0, 0, OCTimg.BMP)
    OCTThicknessLabel.Color = OCTPanel.Color
    OCTThicknessLabel.Caption = ""
END SUB


SUB ClearFovea      'signal init with large negative values
    FovCenter.X = NullResult
    FovCenter.Y = NullResult
    FovCenter.ScreenX = NullResult
    FovCenter.ScreenY = NullResult
END SUB



FUNCTION CheckFovea() AS INTEGER
    RESULT = True
    IF Eye = "R" AND (FovCenter.X > DiskCenter.X) THEN
        RESULT = False
    END IF
    IF Eye = "L" AND (FovCenter.X < DiskCenter.X) THEN
        RESULT = False
    END IF
END FUNCTION



FUNCTION FoveaAngle() AS SINGLE
    IF CheckFovea = False THEN
        RESULT = 0.0!
        EXIT FUNCTION
    END IF

    DEFSNG Retn = 0.0!
    DEFSNG Coeff = IIF(Eye = "R", 1.0, -1.0)
    IF ABS(DiskCenter.X - FovCenter.X) > (4 * SLOmmPerXPixels) THEN
        Retn = Coeff * ATAN( (DiskCenter.Y - FovCenter.Y) / (DiskCenter.X - FovCenter.X)) * Rad2Deg
    ELSE
        Retn = Coeff * 90.0! * SGN(DiskCenter.Y - FovCenter.Y)
    END IF
    RESULT = Retn
END FUNCTION



SUB ResetFovea
    SetDefaultFoveaLocus
    FormRepaint
END SUB



SUB SetFoveaLocus (X AS INTEGER, Y AS INTEGER)
    FovCenter.ScreenX = X
    FovCenter.ScreenY = Y
    FovCenter.X = (X * 1.0!) * SLOmmPerXPixels
    FovCenter.Y = (Y * 1.0!) * SLOmmPerYPixels
    FoveaHasBeenSet = True

    IF CheckFovea = False THEN                                                  'check FovCenter.X is appropriate to DiskCenter.X
        FoveaAngleLabel.Caption = "Fov misaligned (0)"
    ELSE
        FoveaAngleLabel.Caption = "Fov " + STRF$(FoveaAngle(), ffGeneral, 3, 1) + degr
    END IF
END SUB



SUB SetDefaultFoveaLocus
    FovCenter.X = IIF(Eye = "R", DiskCenter.X - 4.9, DiskCenter.X + 4.9)        'estimated standard disk to fovea distance
    FovCenter.Y = DiskCenter.Y + 1.01                                           'this results in fovea angle of 11.6 deg
    FovCenter.ScreenX = (FovCenter.X * 1.0!) / SLOmmPerXPixels
    FovCenter.ScreenY = (FovCenter.Y * 1.0!) / SLOmmPerYPixels
    IF CheckFovea = False THEN                                                  'check FovCenter.X is appropriate to DiskCenter.X
        'StatusLabel.Caption = "Fovea incorrectly aligned"                      'should never be here... just in case
        FoveaAngleLabel.Caption = "Fov misaligned (0)"
    ELSE
        FoveaAngleLabel.Caption = "Fov " + STRF$(FoveaAngle(), ffGeneral, 3, 1) + degr
    END IF
END SUB



SUB DrawFoveaLocus
    IF (FovCenter.ScreenX <= SLOimg.Width) AND (FovCenter.ScreenX > 0) AND _
        (FovCenter.ScreenY <= SLOimg.Height) AND (FovCenter.ScreenY > 0) THEN
        Form.FillRect(FovCenter.ScreenX - 4, FovCenter.ScreenY - 4, FovCenter.ScreenX + 4, FovCenter.ScreenY + 4, RGB(0, 228, 255))
        Form.TextOut(FovCenter.ScreenX, FovCenter.ScreenY -22, "F", FoveaColor, -1)
    END IF
END SUB



FUNCTION CheckDiskSize() AS INTEGER
    RESULT = True
    IF VAL(ScanRadiusEdit.Text) < 100 THEN 
        StatusLabel.Caption = "Radius size too small"                     'set limits here
        ShowMessage "Enter a larger number for scan radius"
        RESULT = False
    END IF
END FUNCTION



SUB SetOpticDiskCenter (X AS INTEGER, Y AS INTEGER)
    DEFSNG Xdif = DiskCenter.ScreenX - X        'get change from prior setting
    DEFSNG Ydif = DiskCenter.ScreenY - Y
    DiskCenter.ScreenX = X
    DiskCenter.ScreenY = Y
    DiskCenter.X = (X * 1.0!) * SLOmmPerXPixels
    DiskCenter.Y = (Y * 1.0!) * SLOmmPerYPixels
    DiscHasBeenSet = True
    'ave is 6.11 degrees +/- 3.32 degrees below center of nerve, ave dist nerve-fovea 4.90 mm, assume 300 um per deg

    IF FovCenter.ScreenX = NullResult THEN 
        SetDefaultFoveaLocus                                        'not set yet, default is ~6 to 12 deg
    ELSE
        IF FoveaHasBeenSet = False THEN                             'shift fovea with the new disk position
        FovCenter.ScreenX = FovCenter.ScreenX - Xdif
        FovCenter.ScreenY = FovCenter.ScreenY - Ydif
        FovCenter.X = (FovCenter.ScreenX * 1.0!) * SLOmmPerXPixels
        FovCenter.Y = (FovCenter.ScreenY * 1.0!) * SLOmmPerYPixels
        END IF
    END IF
END SUB




SUB UpdateScanRadius
    StatusLabel.Caption = "Scan radius = " + STRF$( Circ.Radius * 1000, ffGeneral, 5, 6) + mu
    IF ScanRadiusEdit.Text = "" THEN ScanRadiusEdit.Text = STRF$( Circ.Radius * 1000, ffGeneral, 5, 6)
    Circ.Left   = (Circ.CenterX - Circ.Radius)/ SLOmmPerXPixels
    Circ.Right  = (Circ.CenterX + Circ.Radius) / SLOmmPerXPixels
    Circ.Top    = (Circ.CenterY - Circ.Radius) / SLOmmPerYPixels
    Circ.Bottom = (Circ.CenterY + Circ.Radius) / SLOmmPerYPixels
END SUB



SUB DrawOpticDiskLocus
    IF DiscHasBeenSet THEN
        Form.FillRect(DiskCenter.ScreenX - 4, DiskCenter.ScreenY - 4, DiskCenter.ScreenX + 4, DiskCenter.ScreenY + 4, OpticDiskColor) 
    END IF
END SUB



SUB GetSectionLoci
    DEFINT i = 0
    DEFSNG ang
    DEFSNG TheDist = (1.0! * VAL(ScanRadiusEdit.Text)) / 1000.0!          'convert to HEYEX mm coordinates, hypotenus of right triangle
    StatusLabel.Caption = " "

    IF Line.StartY <> Line.EndY THEN                                'not a horizontal scan
        ShowMessage "Cannot score oblique scans"
        StatusLabel.Caption = "Cannot score oblique scans"
        EXIT SUB
    END IF

    DEFSNG a = DiskCenter.Y - Line.StartY                           'distance (mm) of 1 side in right triangle
    IF ABS(a - TheDist) <= (2*SLOmmPerYPixels) THEN a = TheDist     'clamp to single location resolution
    IF ABS(a) > TheDist THEN                                        'scan is too peripheral
        StatusLabel.Caption = "Scan out of range " + STRF$(a * 1000.0!, ffGeneral, 4, 4) + mu
        EXIT SUB
    END IF

    ClearLoci                                                       'new loci, clear out previous
    'Get line intersection to left side of disk
    DEFSNG b = SQR(TheDist^2 - a^2)                                 'get horiz distance from disk center in mm
    OCTLocus(i).X = DiskCenter.X - b                                'get locus in mm to left of disk center
    ang = ASIN(a/TheDist) * Rad2Deg
    IF (Eye = "R") AND (ang <0.0!) THEN ang = 360.0!+ ang
    IF Eye = "L" THEN ang = ang + 180.0!


    OCTLocus(i).Angle = ang - FoveaAngle                                    'this is corrected angle standard to Heyex coordinates
    IF OCTLocus(i).Angle > 360.00 THEN DEC(OCTLocus(i).Angle, 360.00)       'phase wrap
    IF OCTLocus(i).Angle < 0.00   THEN INC(OCTLocus(i).Angle, 360.00)       'phase wrap
    OCTLocus(i).ScreenX = OCTZoomFactor * OCTLocus(i).X / OCTmmPerXPixels
    OCTLocus(i).SLOscreenX = OCTLocus(i).X / OCTmmPerXPixels                'correspoding horiz location on SLO
    OCTLocus(i).SLOscreenY = Line.Top                                       'correspoding vert location on SLO
    IF (OCTLocus(i).X >= Line.StartX) AND _
       (OCTLocus(i).X <= Line.EndX) AND _
        (a <> TheDist) THEN i++                                             'can add another index

    'Get line intersection to right side of disk
    OCTLocus(i).X = DiskCenter.X + b                                        'get locus in mm to Right of disk center
    IF Eye = "R" THEN 
        ang = 90.0! + ACOS(a/TheDist) * Rad2Deg
    ELSE
        ang = ASIN(a/TheDist) * Rad2Deg
    END IF
    OCTLocus(i).Angle = ang - FoveaAngle                                    'this is final output angle in HEYEX coord
    IF OCTLocus(i).Angle > 360.00 THEN DEC(OCTLocus(i).Angle, 360.00)       'phase wrap
    IF OCTLocus(i).Angle < 0.00   THEN INC(OCTLocus(i).Angle, 360.00)       'phase wrap
    OCTLocus(i).ScreenX = OCTZoomFactor * OCTLocus(i).X / OCTmmPerXPixels
    OCTLocus(i).SLOscreenX = OCTLocus(i).X / OCTmmPerXPixels                'correspoding horiz location on SLO
    OCTLocus(i).SLOscreenY = Line.Top                                       'correspoding vert location on SLO
    IF (OCTLocus(i).X < Line.StartX) OR _
       (OCTLocus(i).X > Line.EndX) THEN i--                                 'can't add this index
    OCTLocus(0).Num = i+1                                                   'how many do we have?
    StatusLabel.Caption = "Number of matches = " + STR$(OCTLocus(0).Num )
END SUB




SUB GetCircleLoci

    DEFINT i, j
    DEFSNG ang
    DEFSNG TheDist = (1.0! * VAL(ScanRadiusEdit.Text)) / 1000.0! 
    StatusLabel.Caption = " "

    ClearLoci                                                                   'new loci, clear out previous segmentation for this image

    'ang of each pixel of OCT is 360 deg/Unscaled OCT Width, can loosen this resolution for speed in search
    'so 360/768 = 0.46875 deg per pixel, or about 1 deg for a faster search with a larger match area
    'for left eye, rotation is CCW starting at the far right (temporal to disk)
    'for right eye, rotation is CW starting at the far left (temporal to disk)

    'disk and desired scan outside of actual scan
    IF SQR((Circ.CenterX- DiskCenter.X)^2 + (Circ.CenterY - DiskCenter.Y)^2) > (TheDist + Circ.Radius) THEN
        StatusLabel.Caption = "Scan out of range " + mu
        EXIT SUB
    END IF

    'desired scan never intersects
    IF SQR((Circ.CenterX- DiskCenter.X)^2 + (Circ.CenterY - DiskCenter.Y)^2) < (Circ.Radius - TheDist) THEN
        StatusLabel.Caption = "No intersection with desired scan"
        EXIT SUB
    END IF

    'This can take time to calculate
    WaitFormLabel.Caption = "Please Wait... " + CHR$(13) + CRLF +_
        "Performing loci matching" + CRLF + "and segmentation"
    WaitForm.Left = OCT_Left + 30      'Form.Width - WaitForm.Width             'get it out of the way
    WaitForm.Top = Form.Top + 30
    WaitForm.Show
    DOEVENTS

    DEFINT indx                                                                 'index of x pixels along actual circle in OCT scan
    DEFINT CanStore                                                             'check for duplicate matches
    DEFSNG SearchAngStep = TwoPi / (1.0! * UnscaledOCTWidth)                    'this one must be constant for 1:1 correspondence
    DEFSNG SearchArea = VAL(SearchResBox.Item(SearchResBox.ItemIndex))          'allow some slop in eye movement during scan
    DEFSNG stepAng = SearchArea * TwoPi / (1.0! * UnscaledOCTWidth)             'step size for desired scan
    DEFSNG SearchAng, SearchX, SearchY                                          'search vars
    DEFSNG LookupX, LookupY                                                     'desired scan loci locations
    DEFSNG MatchDist                                                            'matches based on radii from centers
    DEFSNG CritX = SLOmmPerXPixels * SearchArea                                 'criterion distance for x,y match area
    DEFSNG CritY = SLOmmPerYPixels * SearchArea
    DEFSNG Xcoeff = IIF(Eye = "R", -1.0!, 1.0!)                                 'setup for serial search in HEYEX coordinates

    i = 0
    ang = 0.0!                                                                  'this will be in standard HEYEX angle
    DO
        LookupX = TheDist * Xcoeff * COS(ang) + DiskCenter.X                    'lookup desired scan loci
        LookupY = TheDist * -SIN(ang) + DiskCenter.Y                            'keep consistent for HEYEX angle
        'show the search scan Form.Circle(LookupX /SLOmmPerXPixels, LookupY /SLOmmPerYPixels, 2 + LookupX /SLOmmPerXPixels, 2+LookupY /SLOmmPerYPixels, clYellow, clYellow)

        MatchDist = SQR((Circ.CenterX - LookupX)^2 + (Circ.CenterY - LookupY)^2)
        IF ABS(MatchDist - Circ.Radius) < CritX THEN                            ' AND (ABS(MatchDist - Circ.Radius) < CritY) 
            SearchAng = 0.0!                                                    'dumb algo, do full circle search through scan for match
            indx = 0                                                            'linearly step through the circle scan by pixels
            DO
                SearchX = Circ.Radius * Xcoeff * COS(SearchAng) + Circ.CenterX
                SearchY = Circ.Radius * -SIN(SearchAng) + Circ.CenterY

                IF (ABS(LookupX - SearchX) < CritX) AND (ABS(LookupY - SearchY) < CritY) THEN
                    CanStore = True                                             'search for previous matches
                    FOR j = 0 TO i
                        IF INT(indx * OCTZoomFactor) = OCTLocus(j).ScreenX THEN
                            CanStore = False                                    'already stored this location
                            EXIT FOR
                        END IF
                    NEXT j

                    IF CanStore THEN
                        OCTLocus(i).ScreenX = indx * OCTZoomFactor              'OCTLocus(i).ScreenX = IIF(Eye = "R", indx, UnscaledOCTWidth - indx) 
                        OCTLocus(i).Angle = (ang * Rad2Deg) - FoveaAngle        'this will be output angle with correction for fovea
                        IF OCTLocus(i).Angle > 360.00 THEN DEC(OCTLocus(i).Angle, 360.00)
                        OCTLocus(i).SLOscreenX = SearchX / SLOmmPerXPixels      'correspoding horiz location on SLO
                        OCTLocus(i).SLOscreenY = SearchY / SLOmmPerXPixels      'correspoding vert location on SLO
                        i++
                        INC(SearchAng, SearchAngStep)                           'move to next search, don't over sample
                        indx ++
                        IF i > MaxN THEN
                            i = MaxN                'cannot go more than alloc
                            ang = TwoPi             'force bail out of outer loop
                            EXIT DO
                        END IF
                    END IF
                END IF
                INC(SearchAng, SearchAngStep)
                indx ++
            LOOP UNTIL (indx > UnscaledOCTWidth)    'completed full circle scan
        END IF
        INC(ang, stepAng)
    LOOP UNTIL (ang > TwoPi)                        'completed full re-aligned scan


    IF i > 0 THEN
        IF i = MaxN THEN 
            StatusLabel.Caption = "reached maximum of " + STR$(MaxN) +" matches"
        ELSE
            StatusLabel.Caption = "Number of matches = " + STR$(i)
        END IF
        OCTLocus(0).Num = i                 'how many do we have?
    END IF

    IF OCTLocus(0).Num > 5 THEN             'auto turn on splines if more matches
        SplineButton.Caption = "Lines"
        SplineButton_Click                  'simulate click
    ELSE
        SplineButton.Caption = "Spline"     'auto turn on lines if only a few matches
        SplineButton_Click
    END IF

    WaitForm.Close

END SUB





SUB GetMatchingLoci

    ' ***briefly draw the desired circle, circle coordinates bound the box surrounding the circle ****
    DEFSNG TheDist = (1.0! * VAL(ScanRadiusEdit.Text)) / 1000.0! 
    Form.Circle((DiskCenter.X - TheDist) /SLOmmPerXPixels, (DiskCenter.Y - TheDist) /SLOmmPerYPixels, (DiskCenter.X + TheDist) /SLOmmPerXPixels, (DiskCenter.Y + TheDist)  /SLOmmPerYPixels, clYellow, -1)


    IF ScanType = "SECTION" THEN GetSectionLoci             'just a few matches are possible, quick
    IF ScanType = "CIRCLE" THEN GetCircleLoci               'recalc with either new fovea or disk location

    DrawOCTSegmentation
    ShowSplineControlPoints (False)
    'IF DiscHasBeenSet THEN GetOpticDiscMovement            'in theory measures disc shear / movement
    FormRepaint

    IF NormForm.Visible THEN
        NormCalculate
        NormForm_Paint
    END IF
    WaitForm.Close

END SUB





SUB DrawSLOScan

    'IF OCTPositionTolerance = "N" THEN SLOimg.PenStyle = 1 ELSE SLOimg.PenStyle = 0     'PS_SOLID_BMP=0 , PS_DASH_BMP=1 
    IF ScanType = "SECTION" THEN
        SLOimg.Line(Line.Left, Line.Top, Line.Right, Line.Top, clGreen)
    ELSEIF ScanType = "CIRCLE" THEN
        SLOimg.Circle(Circ.Left, Circ.Top, Circ.Right, Circ.Bottom, clGreen)', -1)      'circle parameters xleft, ytop, xRight, yBottom, color, fill
    END IF



    'start test code eye movement artifact simulation *****************************************************************
    'FormRepaint
    'DIM fDialog     AS QOPENDIALOG
    'DIM InputList   AS QSTRINGLIST
    'InputList.Clear
    'fDialog.Caption = "Load an Eye Movement trace"
    'fDialog.filter = "text files|*.txt;*.csv|config files|*.bin;|All Files|*.*"
    'fDialog.InitialDir = "C:\Rapidq\Proj\VOG_analysis\TestData"
    'fDialog.FileName = ""
    'IF (fDialog.Execute) THEN
    'InputList.LoadFromFile(fDialog.FileName)
    '
    'DEFINT i = 0
    'WHILE (LEFT$(InputList.Item(i), 1) = ";" )
    'i++
    'WEND
    '
    '
    'DEFSTR delimChar = ","                                                              'Comma is default delimiter 
    'IF INSTR(InputList.Item(i), delimChar) = 0 THEN delimChar = CHR$(9)                 'set delimiter to tab if not found
    'IF INSTR(InputList.Item(i), delimChar) = 0 THEN delimChar = CHR$(32)                'set delimiter to space if not found
    'DEFINT myNumColumns = TALLY(InputList.Item(i), delimChar) + 1                       'how many columns of data?
    'DEFSNG TimeT
    'DEFSNG Horiz, dH
    'DEFSNG Vert, dV
    'DEFINT intens
    '
    'DO
    'TimeT = VAL(FIELD$(InputList.Item(i), delimChar, 1))                          'Time[s]	
    'Horiz = (VAL(FIELD$(InputList.Item(i), delimChar, 2)) * 0.290) / SLOmmPerXPixels   'right hor.[] to mm to pixels
    'Vert = (VAL(FIELD$(InputList.Item(i), delimChar, 3))  * 0.290) / SLOmmPerYPixels  'right ver.[] to mm to pixels
    'i++                                                         'inc input file line number
    '
    '
    'intens = 255 - ((SQR( (Horiz - dH)^2 + (Vert - dV)^2) ) * 15)
    ''d.print intens
    '
    'IF ScanType = "SECTION" THEN
    'SLOimg.Line(Line.Left + Horiz, Line.Top + Vert, Line.Right + Horiz, Line.Top + Vert, RGB(0, intens, 0))
    'ELSEIF ScanType = "CIRCLE" THEN
    'SLOimg.Circle(Circ.Left + Horiz, Circ.Top + Vert, Circ.Right + Horiz, Circ.Bottom + Vert, RGB(0, intens, 0))', -1)      'circle parameters xleft, ytop, xRight, yBottom, color, fill
    'END IF
    'dH = Horiz
    'dV = Vert
    '
    'LOOP UNTIL (i = InputList.ItemCount) OR (i > 12000)
    'EXIT SUB
    'END IF

    'end test code *****************************************************************


END SUB




SUB InvertOCT
    OCTImg.InvertColor(0, 0, OCTImg.Width, OCTImg.Height)
    DrawOCTSegmentation
    FormRepaint
END SUB




SUB DrawLociOnSLO
    IF OCTLocus(0).Num > 0 THEN
        DEFINT i
        FOR i = 0 TO OCTLocus(0).Num -1                    'how many do we have? won't plot if num = 0
            Form.Circle(OCTLocus(i).SLOscreenX - 4, OCTLocus(i).SLOscreenY  - 4, _
                        OCTLocus(i).SLOscreenX + 4, OCTLocus(i).SLOscreenY  + 4, clGreen, clRed)
 
        NEXT i
    END IF
END SUB




SUB OCTMatchesVisibility_OnChange
    IF OCTLocus(0).Num > 0 THEN DrawMatchingSegments
    DrawSplines
    FormRepaint
END SUB





SUB DrawMatchingSegments
    DEFINT i, x, y
    TweakOCT.Draw(0, 0, OCTimg.BMP)

    IF SplineButton.Tag THEN

        'blend the lines into the OCT image
        LinesBitmap.Width = TweakOCT.Width
        LinesBitmap.Height = TweakOCT.Height
        LinesBitmap.FillRect(0, 0, LinesBitmap.Width, LinesBitmap.Height, &HFFFFFF)

        DEFINT TheClr =          IIF((MatchSegmentClr AND &H0000FF) = &H0000FF, &H0000FF, (255 - (OCTMatchesVisibilityTrackbar.Position * 2)))
               TheClr = TheClr + IIF((MatchSegmentClr AND &H00FF00) = &H00FF00, &H00FF00, (255 - (OCTMatchesVisibilityTrackbar.Position * 2)) SHL 8)
               TheClr = TheClr + IIF((MatchSegmentClr AND &HFF0000) = &HFF0000, &HFF0000, (255 - (OCTMatchesVisibilityTrackbar.Position * 2)) SHL 16)

        FOR i = 0 TO OCTLocus(0).Num -1                    'how many do we have? won't plot if num = 0
            x = OCTLocus(i).ScreenX
            LinesBitmap.Line(x, 0, x, LinesBitmap.Height, TheClr)
        NEXT i
        TweakOCT.CopyMode = cmSrcAnd
        TweakOCT.Draw(0, 0, LinesBitmap.BMP)
        TweakOCT.CopyMode = cmSrcCopy

    ELSE

        TweakOCT.PenSize = 1 '4
        FOR i = 0 TO OCTLocus(0).Num -1                    'how many do we have? won't plot if num = 0
            x = OCTLocus(i).ScreenX
            TweakOCT.Line(x, 0, x, TweakOCT.Height, MatchSegmentClr)
            IF OCTLocus(i).ScreenThickness > 0 THEN
                TweakOCT.Line(x - 4, OCTLocus(i).ScreenY, x + 4, OCTLocus(i).ScreenY, clPurple)
                TweakOCT.Line(x - 4, OCTLocus(i).ScreenThickness, x + 4, OCTLocus(i).ScreenThickness, clPurple)
                TweakOCT.Line(x, OCTLocus(i).ScreenY, x, OCTLocus(i).ScreenThickness, clPurple)
            END IF
        NEXT i
        TweakOCT.PenSize = 2  '4

    END IF
END SUB




SUB DrawOCTSegmentation
    TweakOCT.Draw(0, 0, OCTimg.BMP)
    IF OCTLocus(0).Num > 0 THEN DrawMatchingSegments
    IF SplineButton.Tag THEN DrawSplines
END SUB





SUB FormRepaint
    '*****  SLO image on left **********
    IF SLOimg.Width > 0 THEN
        Form.Draw(0, 0 , SLOimg.BMP)            'keep on left side for form coordinates
        DrawOpticDiskLocus                      'center of optic disk
        DrawFoveaLocus                          'center of fovea
        IF (MyImageMatch.centerX <> 0) AND (MyImageMatch.centerY <> 0) THEN DrawDiscMatch           'IF (ReferenceDiscBMP.Width <> ReferenceImageWidth) THEN   'reference set '
    END IF

    '*****  OCT image on Right **********
    IF OCTimg.Width > 0 THEN 
        Form.Draw(OCT_Left, 0, TweakOCT.BMP)
        DrawLociOnSLO
        'form.draw(0, 0, EdgeBMP.BMP)
        'form.draw(0, 200, kMeanCopy.BMP)
        'form.draw(0, 400, VesselMask.BMP)
    END IF
END SUB




SUB LinesMouseDown (Button AS INTEGER, X AS INTEGER, Y AS INTEGER, Shift AS INTEGER)
    SelectedIndx = -1                                               'flag not found

    IF OCTLocus(0).Num > 0 THEN
        DEFINT i
        FOR i = 0 TO OCTLocus(0).Num            'MaxN
            IF ABS(X - OCT_Left - OCTLocus(i).ScreenX) < 3 THEN     'close to plotted line by 3 pixels
                SelectedIndx = i                                    'store for drag and release
                OCTLocus(SelectedIndx).ScreenThickness = 0
                OCTLocus(SelectedIndx).ScreenY = Y                  'store location of start click
                DrawMatchingSegments
                EXIT FOR
            END IF
        NEXT i
    END IF
END SUB




SUB LinesMouseUp (Button AS INTEGER, X AS INTEGER, Y AS INTEGER, Shift AS INTEGER)
    IF SelectedIndx > -1 THEN
        'X = OCTLocus(SelectedIndx).ScreenX
        OCTLocus(SelectedIndx).ScreenThickness = Y
        OCTLocus(SelectedIndx).RNFL_Thickness = ABS(Y - OCTLocus(SelectedIndx).ScreenY) * 1000.0! * OCTmmPerYPixels / OCTZoomFactor  'store thickness from start-end click in microns
        SelectedIndx = -1
        DrawMatchingSegments
    END IF
END SUB




SUB LinesMouseMove (x AS INTEGER, y AS INTEGER, Shift AS INTEGER)
    IF SelectedIndx > -1 THEN
        Form.Draw(OCT_Left, 0, TweakOCT.BMP) 
        x = OCT_Left + OCTLocus(SelectedIndx).ScreenX
        OCTLocus(SelectedIndx).ScreenThickness = y

        Form.Line(x - 4, OCTLocus(SelectedIndx).ScreenY, _
                  x + 4, OCTLocus(SelectedIndx).ScreenY, clPurple)      'start tick
        Form.Line(x, OCTLocus(SelectedIndx).ScreenY, x, y, clPurple)    'connecting line
        Form.Line(x - 4, y, x + 4, y, clPurple)                         'end tick
        DEFSTR CurThick = STRF$( ABS(Y - OCTLocus(SelectedIndx).ScreenY)* 1000.0 * OCTmmPerYPixels / OCTZoomFactor ,  ffGeneral, 5, 6) + mu
        OCTThicknessLabel.Color = 0
        OCTThicknessLabel.Caption = CurThick            'Form.TextOut(SLOImg.Width - 70, SLOImg.Height + 20, CurThick, clGreen, 0)
    END IF
END SUB



'--------------------------------------------------------------------
'   search in OCT image if mouse click is near a segmentation line
'--------------------------------------------------------------------

FUNCTION GetLayerIsPicked(x AS INTEGER, y AS INTEGER) AS INTEGER
    DEFINT i, j, k
    DEFINT Xsrch = 8, Ysrch = 8                                                     'search range +/-
    DEFINT ActiveLayer = cNone                                                      'return matched layer from segmentation line

    FOR i = 0 TO Xsrch                                                              'search horizontal 
        FOR j = 0 TO Ysrch                                                          'search vertical on inner loop to best find vertical match
            FOR k = 0 TO MaxLayers
                IF (TweakOCT.Pixel(x-i-OCT_Left, y-j) = SegColor(k)) THEN           'symmetric search from middle of dblClick location
                    ActiveLayer = k
                    EXIT FOR
                END IF
                IF (TweakOCT.Pixel(x+i-OCT_Left, y+j) = SegColor(k)) THEN           'symmetric search from middle of dblClick location
                    ActiveLayer = k
                    EXIT FOR
                END IF
            NEXT k
            IF ActiveLayer <> cNone THEN EXIT FOR
        NEXT j
        IF ActiveLayer <> cNone THEN EXIT FOR
    NEXT i

    RESULT = ActiveLayer
END FUNCTION






SUB Form_MouseDown(Button AS INTEGER, X AS INTEGER, Y AS INTEGER, Shift AS INTEGER)

    IF (X <= SLOimg.Width) AND (Y <= SLOimg.Height) THEN
        '************************************************
        '******   mouse clicked in SLO image  ***********
        '************************************************

        SLOflag = True

        IF Button = MouseRtButton THEN                                  'right button click on SLO

            IF Shift THEN                                               '(Shift = 16) for Alt key down
                WindowLevel.X = True
                WindowLevel.ScreenX = X
                WindowLevel.ScreenY = Y
                EXIT SUB                                                'go set contrast / brightness in mouse move
            ELSE
                SetFoveaLocus (X, Y)                                    'Right mouse click sets fovea location
                IF NormForm.Visible THEN NormBMPClear
            END IF

        ELSE                                                            'left button click on SLO
            IF CheckDiskSize = False THEN EXIT SUB
            SetOpticDiskCenter (X, Y)
        END IF

        IF DiscHasBeenSet THEN GetMatchingLoci


    ELSEIF (X >= OCT_Left) AND (Y <= OCTimg.Height) THEN

        '*************************************************
        '*********** clicked in OCT image ****************
        '*************************************************

        IF (Button = MouseLtButton) THEN

            IF ((TIMER - LastTimeMouseDown) < 0.3) THEN                     'DblClick of left button?
                IF CptN(CurrLayer) > 1 THEN
                    LastTimeMouseDown = TIMER                               'reset for lag to follow
                    AddControlPointbyDblClick(X, Y)                         'dblClick on form to add control point
                    GetSpline                                               'update segmentation
                    DrawOCTSegmentation
                    FormRepaint
                END IF

            ELSE                                                            'single left button click

                IF SplineButton.Tag = True THEN                             'using splines to measure retinal thickness
                    SplineMouseDown(Button, X, Y, Shift)
                ELSE                                                        'using individual lines to measure retinal thickness
                    LinesMouseDown(Button, X, Y, Shift)
                END IF
                LastTimeMouseDown = TIMER
            END IF

        ELSE                                                                

            '  default Right mouse button down
            '****user wants to adjusts window level******

            WindowLevel.X = True
            WindowLevel.ScreenX = X
            WindowLevel.ScreenY = Y                                     'go set contrast / brightness in mouse move
            SLOflag = False                                             'signal change OCT image
            LastTimeMouseDown = TIMER
        END IF

    END IF
END SUB




SUB Form_MouseUp(Button AS INTEGER, X AS INTEGER, Y AS INTEGER, Shift AS INTEGER)

    IF WindowLevel.X THEN                               'flag ****user adjusting window level******
        WindowLevel.X = False
        IF SLOflag THEN
            SLOImg.Draw(0, 0, TempContrastBMP.BMP)           'should already same size & format
            DrawSLOScan
            FormRepaint
        ELSE
            OCTImg.Draw(0, 0, TempContrastBMP.BMP)           'should already same size & format
            DoSegmentation
        END IF
        SLOflag = False
        EXIT SUB
    END IF

    '**** else user adjusts RNFL thickness end point ******
    IF (X >= OCT_Left) AND (Y <= OCTimg.Height) THEN        'release click in OCT image
        IF SplineButton.Tag THEN                            'splines turned on
            SplineMouseUp(Button, X, Y, Shift)
        ELSE
            LinesMouseUp(Button, X, Y, Shift)
        END IF
    END IF
    FormRepaint
END SUB




SUB Form_MouseMove( X AS INTEGER, Y AS INTEGER, Shift AS INTEGER)

    STATIC ImBusy
    IF ImBusy = False THEN                                  'check events are building up when too slow!!!

        IF WindowLevel.X THEN                               'flag ****user adjusting window level******
            ImBusy = True
            'real time update of effect if image selected store in SegmentBMP as TEMPORARY bmp
            IF SLOflag THEN
                TempContrastBMP.Width = SLO_Orig.Width
                TempContrastBMP.Height = SLO_Orig.Height
                TempContrastBMP.PixelFormat = pf8bit
                TempContrastBMP.Draw(0, 0, SLO_Orig.BMP)
            ELSE
                TempContrastBMP.Width = OCTimg.Width
                TempContrastBMP.Height = OCTimg.Height
                TempContrastBMP.PixelFormat = pf8bit
                TempContrastBMP.Draw(0, 0, OCTimg.BMP)
            END IF

            'set window levels with mouse
            cv_Bitmap2Mat(TempContrastBMP.Width, TempContrastBMP.Height, 1, TempContrastBMP.Pointer)        'first store it in OpenCV interface

            DEFDBL alpha = 1.0! + (1.5! * (X - WindowLevel.ScreenX) /  TempContrastBMP.Width)               'place contrast value in double var for pointer   
                    IF alpha < 0.05 THEN alpha = 0.05
                    IF alpha > 4.00 THEN alpha = 4.00

            DEFDBL beta = 0.0# + (1.0# * (WindowLevel.ScreenY - Y) / 10.0#)
                    IF beta < -160 THEN beta = -160
                    IF beta > 160 THEN beta = 160

            cv_ContrastWindow(alpha , beta)


            'real time update
            IF SLOflag THEN
                Form.Draw(0, 0, TempContrastBMP.BMP)
            ELSE
                Form.Draw(OCT_Left, 0, TempContrastBMP.BMP)
            END IF
            DOEVENTS

            ImBusy = False
            EXIT SUB
        END IF

        '**** else user adjusting RNFL thickness end point ******
        IF (X >= OCT_Left) AND (Y <= OCTimg.Height) THEN            'release click in OCT image
            ImBusy = True

            IF SplineButton.Tag THEN                                'splines turned on
                IF CtrlKeyIsDown THEN                               'move control points by hovering mouse
                    ShowSplineControlPoints (True)                  'turn on just in case not selected yet
                    MoveControlPointByCircle(X, Y)
                ELSE
                    SplineMouseMove(X, Y, Shift)
                END IF
            ELSE
                LinesMouseMove(X, Y, Shift)
            END IF

            ImBusy = False
        END IF

    END IF
END SUB



'------------------------------------------------------------------------------------------
'   median filter button clicked
'------------------------------------------------------------------------------------------

SUB MedianFilter_Click
    IF MedianFilterBox.Checked = False THEN EXIT SUB

    STATIC ImBusy
    IF ImBusy = False THEN
        ImBusy = True
        DEFINT MedSz = VAL(MedianFilterSizeBox.Item(MedianFilterSizeBox.ItemIndex))
        WaitFormLabel.Caption = "Please Wait... " + CHR$(13) + CRLF +_
            "Performing median filtering" + CRLF + _
            "size " + STR$(MedSz) + " x " + STR$(MedSz) + "pixels."
        WaitForm.Center
        WaitForm.Show
        DOEVENTS

        'update working bitmap
        TempContrastBMP.Width = OCTimg.Width
        TempContrastBMP.Height = OCTimg.Height
        TempContrastBMP.Draw(0, 0, OCTimg.BMP)
        cv_Bitmap2Mat(TempContrastBMP.Width, TempContrastBMP.Height, 1, TempContrastBMP.Pointer)    'first store it in OpenCV interface
        cv_MedianBlur(MedSz)
        OCTimg.Draw(0, 0, TempContrastBMP.BMP)
        DoSegmentation
        WaitForm.Close
        DOEVENTS
        ImBusy = False
    END IF
END SUB




'------------------------------------------------------------------------------------------
'   Revert button clicked
'------------------------------------------------------------------------------------------

SUB ResetOCTImage
    OCTimg.BMP = OCT_Orig.BMP
    DoSegmentation
END SUB



SUB ResetSLOimage
    SLOimg.BMP = SLO_Orig.BMP
    DrawSLOScan
    FormRepaint
END SUB



SUB LoadOCTImage

    TweakOCT.Width = 1
    TweakOCT.Height = 1
    TweakOCT.LoadFromFile(OCTFName)                     'use for temporary loading

    IF (TweakOCT.Width <= 1) OR (TweakOCT.Height <= 1) THEN
        Showmessage "Unknown image load error" + CRLF +_
                    "Please reload image"
        EXIT SUB
    END IF

    MyRect.Right = TweakOCT.Width * OCTZoomFactor       'set zoom factor
    MyRect.Bottom = TweakOCT.Height * OCTZoomFactor

    OCT_Orig.Width = MyRect.Right                       'resize final output
    OCT_Orig.Height = MyRect.Bottom
    OCT_Orig.PixelFormat = pf24bit
    OCT_Orig.StretchDraw(MyRect, TweakOCT.BMP)          'store it

    OCTimg.Width = MyRect.Right                         'resize final output
    OCTimg.Height = MyRect.Bottom
    OCTimg.PixelFormat = pf8bit                         'default DC is 32-bit, forget alpha channel
    OCTimg.GreyScale
    OCTimg.Draw(0, 0, OCT_Orig.BMP)                     'make our copy of original

    'Tweaked OCT has user modified image (contrast, gamma, brightness, median)
    'this will be loaded up after filtering / segmentation
    TweakOCT.Width = OCTimg.Width
    TweakOCT.Height = OCTimg.Height
    TweakOCT.PixelFormat = pf24bit
    TweakOCT.Draw(0, 0, OCTimg.BMP)

END SUB





SUB Form_KeyUp(Key AS WORD, Shift AS INTEGER)
    IF CtrlKeyIsDown THEN
        CtrlKeyIsDown = False
        ControlPointCursorSGN = NullResult
        Form.Cursor = crDefault
    END IF

    DeleteKeyDown = False
END SUB



SUB Form_KeyDown(Key AS WORD , Shift AS INTEGER)

    IF (Shift = 1) AND (key = 17)  THEN             'Ctrl only key
        CtrlKeyIsDown = True
        IF SplineButton.Tag THEN Form.Cursor = crSize 
        EXIT SUB
    END IF 

    IF (Shift = 16) AND (key = 18)  THEN            'Alt only key, same as ctrl key
        CtrlKeyIsDown = True
        IF SplineButton.Tag THEN Form.Cursor = crSize 
        EXIT SUB
    END IF 

    IF (Shift = 1) AND (key = 66) THEN              'ctrl-b key
        UseVesselMaskCheckBox.Checked = IIF(UseVesselMaskCheckBox.Checked, False, True)
        DoSegmentation
        EXIT SUB
    END IF

    IF key = 46 THEN                                'del key 
        DeleteKeyDown = True
        EXIT SUB
    END IF

    IF key = 77 THEN                                'm key
        DEFINT PriorCheck = MedianFilterBox.Checked
        MedianFilterBox.Checked = True              'auto generates MedianFilter_Click
        MedianFilterBox.Checked = PriorCheck
        'MedianFilterBox.Checked = IIF(MedianFilterBox.Checked, False, True)
        'IF MedianFilterBox.Checked THEN MedianFilter_Click
    END IF

    IF key = 65 THEN                                'a key
        AdaptiveThresholdCheckBox.Checked = IIF(AdaptiveThresholdCheckBox.Checked, False, True)
        DoSegmentation
    END IF

    IF key = 66 THEN                                'b key
        CurrLayer = cBM                             'activate BM layer
        IF SplineButton.Tag THEN ShowSplineControlPoints (True)
        DrawOCTSegmentation                             'Draws splines
    END IF

    IF key = 73 THEN                                'i key
        CurrLayer = cILM                            'activate ILM layer
        IF SplineButton.Tag THEN ShowSplineControlPoints (True)
        DrawOCTSegmentation                             'Draws splines
    END IF

    IF key = 78 THEN                                'n key
        CurrLayer = cNFL                            'activate RNFL layer
        IF SplineButton.Tag THEN ShowSplineControlPoints (True)
        DrawOCTSegmentation                         'Draws splines
    END IF

    IF key = 80 THEN                                'p key, polynomial curve fit
        PolyFitLayersCheckBox.Checked = IIF(PolyFitLayersCheckBox.Checked, False, True)
        DoSegmentation
    END IF

    IF key = 82 THEN InvertOCT                      'r key

    IF key = 86 THEN                                'v key
        IF (Shift <> 0) THEN                        'shift-v or ctrl-v or alt-v key
            ViewReference
        ELSE
            ToggleNormTemplate
        END IF
    END IF


    'Screen Capture
    'IF key = 80 THEN  'p
    'DEFLNG  hdc = getdc(form.Handle)
    'myBMP.Width = form.ClientWidth
    'myBMP.Height = form.ClientHeight
    'BitBlt(myBMP.Handle , 0 , 0 , form.ClientWidth , form.ClientHeight , hdc ,  , 0 , &HCC0020)
    'myBMP.SaveToFile "untitled.bmp"
    'releasedc(form.Handle , hdc)
    'END IF

END SUB





SUB ChangeDirectories
    ChDir(DirList.Item(DirList.ItemIndex) - "[" - "]")
    DirList.Directory = CurDir$
    XMLlist.Directory = CurDir$
    CurrentFolderNameLabel.Caption = XMLlist.Directory
END SUB



SUB CancelOpenFolderForm
    CancelFolderButton.Tag = True
    OpenFolderForm.Close
END SUB



SUB OpenXMLFolder
    CancelFolderButton.Tag = False
    OpenFolderForm.Close
END SUB



SUB OpenFolderFormResize
    CancelFolderButton.Top = OpenFolderForm.ClientHeight - CancelFolderButton.Height - 8
    SelectFolderButton.Top = OpenFolderForm.ClientHeight - SelectFolderButton.Height - 8
    SelectFolderButton.Left = OpenFolderForm.ClientWidth - SelectFolderButton.Width - 8

    CurrentFolderNameLabel.Width = OpenFolderForm.ClientWidth - 8
    XMLlist.Top = CurrentFolderNameLabel.Top + CurrentFolderNameLabel.Height + 8
    XMLlist.Left = 12 + OpenFolderForm.ClientWidth \2
    XMLlist.Height = OpenFolderForm.ClientHeight - CurrentFolderNameLabel.Height - CancelFolderButton.Height - 22
    XMLlist.Width = OpenFolderForm.ClientWidth \2 - 18

    DirList.Top = CurrentFolderNameLabel.Top + CurrentFolderNameLabel.Height + 8
    DirList.Height = OpenFolderForm.ClientHeight - CurrentFolderNameLabel.Height - CancelFolderButton.Height - 22
    DirList.Width = OpenFolderForm.ClientWidth \2
END SUB




FUNCTION GetXMLDataFolder() AS INTEGER
'    CancelFolderButton.Tag = False
'    DO
'        OpenFolderForm.Show
'        DOEVENTS
'    LOOP UNTIL (OpenFolderForm.Visible = False)
'    IF CancelFolderButton.Tag = False THEN
'        ChDir XMLlist.Directory
'        RESULT = True
'    ELSE
'        RESULT = False
'    END IF
    MyOpenDialog.Filter = "XML files|*.xml|*.*"
    MyOpenDialog.FilterIndex = 1

    IF MyOpenDialog.Execute THEN
        RESULT = True
    ELSE
        RESULT = False
    END IF


END FUNCTION




'--------------------------------------------------------------------------------------------
' find exported filename starts with file:///C:xxxx, we may be loading file from external folder or moved images to app folder
'--------------------------------------------------------------------------------------------

SUB FindHeyexFile

    IF LEFT$(myFileName, 8) = "FILE:///" THEN 
        myFileName = RIGHT$(myFileName, LEN(myFileName) -8)
    END IF

    IF FILEEXISTS(myFileName) = False THEN                                  'not where Heyex last left it
        myFileName = StripFileName(myFileName)                              'get basic file name
        myFileName = Application.Path + "\" + myFileName                    'try copied to app folder
        IF FILEEXISTS(myFileName) = False THEN 
            myFileName = StripFileName(myFileName)                          'still cant find it try default path
            IF FILEEXISTS(myFileName) = False THEN                          'I give up
                Showmessage "Error finding file " + myFileName + ". Exiting"
                DEFINT Fini = True
                Form_Close(Fini)
            END IF
        END IF
    END IF
END SUB



'--------------------------------------------------------------------------------------------
'move current string pointer to an XML field, does not search for sub tags
'--------------------------------------------------------------------------------------------
FUNCTION LocateXMLTag(theTag AS STRING, theTxt AS QSTRINGLIST, stPos AS INTEGER) AS INTEGER
    DEFSTR myTxt = UCASE$(TheTxt.Text)
    DEFSTR myTag = "<" + UCASE$(TheTag) + ">"
    DEFINT i = 0

    i = INSTR(stPos, myTxt, myTag)
    IF i > 0 THEN RESULT = i + LEN(myTag) -1 ELSE RESULT = stPos
END FUNCTION




'--------------------------------------------------------------------------------------------
'Returns a single simple XML field in uppercase, does not search for sub tags
'--------------------------------------------------------------------------------------------
FUNCTION GetXMLValue(theTag AS STRING, theTxt AS QSTRINGLIST, ByRef stPos AS INTEGER) AS STRING
    DEFSTR myTxt = UCASE$(TheTxt.Text)
    DEFSTR myTag = "<" + UCASE$(TheTag) + ">"
    DEFSTR myEndTag = "</" + UCASE$(TheTag) + ">"
    DEFSTR myEndField = "</"
    DEFINT i = 0, j = 0

    RESULT = ""
    i = INSTR(stPos, myTxt, myTag) 
    IF i > 0 THEN
        j = INSTR(i + LEN(myTag), myTxt, myEndField)         'get end of field
        RESULT = MID$(MyTxt, i + LEN(myTag), j-i-LEN(myTag))
        stPos = j + LEN(myEndField)
    END IF
END FUNCTION







'***********************************************************
'  Main loop, run through all files in current dir
'***********************************************************

Form.Show



DO
    OutputFileName = ""                 'start a new output file
    MyDir = CURDIR$
    ClearAllData                        'clear out existing memory, reset selections
    myFileName = DIR$("*.xml", 0)       'get HEYEX xml file

    IF myFileName = "" THEN
        Form.TextOut(40, 80,  "  There are no exported files found in the current folder.  ", clRed, clBlack)
        Form.TextOut(40, 110, "  Please choose the folder with the exported XML files.     ", clRed, clBlack)

        IF (GetXMLDataFolder = False) THEN Application.Terminate
        myFileName = DIR$("*.xml", 0)       'get HEYEX xml file
    END IF


    WHILE  myFileName <> ""
        IF FILEEXISTS(myFileName) THEN
            sList.LoadFromFile(myFileName)
            ClearFovea
            sPos = 0
            MRN =   GetXMLValue("PatientID", sList, sPos)
            LName = GetXMLValue("LastName", sList, sPos)
            FName = GetXMLValue("FirstNames", sList, sPos)
            NameLabel.Caption = "Name: " + FName + ", " + LName + ",  " + MRN
            cPos = sPos

            'each subject may have many series in one XML file
            WHILE (LocateXMLTag("Series", sList, sPos) > cPos)              'one set of SLO/OCT is in one series
                ScanType = UCASE$(GetXMLValue("Type", sList, sPos))         'series type is circle or section

                sPos = LocateXMLTag("Image", sList, sPos)                   'get SLO image
                Eye = GetXMLValue("Laterality", sList, sPos)
                EyeLabel.Caption = "Eye: " + Eye
                sPos = LocateXMLTag("ImageType", sList, sPos)
                ImgType = GetXMLValue("Type", sList, sPos)


                IF ImgType = "LOCALIZER" THEN

                    SLOimg.Width = VAL(GetXMLValue("Width", sList, sPos))
                    SLOimg.Height = VAL(GetXMLValue("Height", sList, sPos))
                    SLOmmPerXPixels = VAL(GetXMLValue("ScaleX", sList, sPos))
                    SLOmmPerYPixels = VAL(GetXMLValue("ScaleY", sList, sPos))
                    SLOFocus = GetXMLValue("Focus", sList, sPos)
                    myFileName = GetXMLValue("ExamURL", sList, sPos)
                    FindHeyexFile                                       'sets myFileName to proper location
                    SLOimg.LoadFromFile(myFileName)
                    SLO_Orig.Width = SLOimg.Width
                    SLO_Orig.Height = SLOimg.Height
                    SLO_Orig.PixelFormat = pf24bit                      'default DC is 32-bit, forget alpha channel
                    SLO_Orig.Draw(0, 0, SLOimg.BMP)                     'Make a backup
                    OCT_Left = SLOimg.Width + 2                         'smallest gap possible
                ELSE

                    Showmessage "ERROR in XML sequence!!"

                END IF


                sPos = LocateXMLTag("Image", sList, sPos)
                IF (Eye <> GetXMLValue("Laterality", sList, sPos)) THEN Showmessage "ERROR in Eye sequence!!"
                sPos = LocateXMLTag("ImageType", sList, sPos)
                ImgType = GetXMLValue("Type", sList, sPos)



                IF ImgType = "OCT" THEN

                    UnscaledOCTWidth = VAL(GetXMLValue("Width", sList, sPos))
                    OCTimg.Width = UnscaledOCTWidth
                    UnscaledOCTHeight = VAL(GetXMLValue("Height", sList, sPos))
                    OCTimg.Height = UnscaledOCTHeight
                    OCTmmPerXPixels = VAL(GetXMLValue("ScaleX", sList, sPos))
                    OCTmmPerYPixels = VAL(GetXMLValue("ScaleY", sList, sPos))
                    ARTaverage = GetXMLValue("NumAve", sList, sPos)
                    OCTQuality = GetXMLValue("ImageQuality", sList, sPos)
                    OCTPositionTolerance = UCASE$(GetXMLValue("PositionWithinTolerance", sList, sPos))
                    OCTPositionTolerance = IIF(OCTPositionTolerance = "FALSE", "N", "Y")
                    QualityLabel.Caption = "Quality: " + OCTQuality + " ("+ ARTaverage + ": " + OCTPositionTolerance + ")"

                    IF ScanType = "SECTION" THEN

                        sPos = LocateXMLTag("Start", sList, sPos)
                            Line.StartX = VAL(GetXMLValue("X", sList, sPos))
                            Line.Left   = Line.StartX / SLOmmPerXPixels
                            Line.StartY = VAL(GetXMLValue("Y", sList, sPos))
                            Line.Top    = Line.StartY / SLOmmPerXPixels
                        sPos = LocateXMLTag("End", sList, sPos)
                            Line.EndX   = VAL(GetXMLValue("X", sList, sPos))
                            Line.Right  = Line.EndX / SLOmmPerXPixels
                            Line.EndY   = VAL(GetXMLValue("Y", sList, sPos))
                            Line.Bottom = Line.EndY / SLOmmPerXPixels
                        StatusLabel.Caption = "Scan length = " + STRF$((Line.EndX - Line.StartX) * 1000, ffGeneral, 5, 6) + mu

                    ELSEIF ScanType = "CIRCLE" THEN

                        sPos = LocateXMLTag("Start", sList, sPos)
                            Circ.StartX = VAL(GetXMLValue("X", sList, sPos))        'these values are not modified
                            Circ.StartY = VAL(GetXMLValue("Y", sList, sPos))
                        sPos = LocateXMLTag("Center", sList, sPos)
                            Circ.CenterX   = VAL(GetXMLValue("X", sList, sPos))
                            Circ.CenterY   = VAL(GetXMLValue("Y", sList, sPos))
                        tmp = UCASE$(GetXMLValue("Direction", sList, sPos))
                            IF tmp = "CCW" THEN         'for left eye, start is to the far right
                                IF Eye <> "L" THEN Showmessage "ERROR, wrong eye in XML file"
                                Circ.Dir = -1 
                                Circ.Radius = Circ.StartX - Circ.CenterX
                            ELSE                        'for right eye, start is to the far left
                                IF Eye <> "R" THEN Showmessage "ERROR, wrong eye in XML file"
                                Circ.Dir = 1         'left eye uses CCW, OD = CW
                                Circ.Radius = Circ.CenterX - Circ.StartX
                            END IF
                            MEMCPY (UDTPTR(OrigCirc), UDTPTR(Circ), SIZEOF(Circ))               'copy the orignal values, in case setting reference 
                            UpdateScanRadius
                    ELSE

                        Showmessage "ERROR, unknown Scan type in XML "

                    END IF

                    myFileName = GetXMLValue("ExamURL", sList, sPos)
                    FindHeyexFile
                    OCTFName = myFileName                               'keep backup name for revert
                    LoadOCTImage
                    OCTPanel.Top = OCTimg.Height + 12                   'relocate panel to bottom of oct image

                    IF ScanType = "SECTION" THEN                        'auto turn on lines for section scans

                        SplineButton.Caption = "Spline"                 'this will toggle off splines to line
                        SplineButton_Click                              'simulate click
                    ELSE

                        SplineButton.Caption = "Lines"                  'this will toggle off lines to splines
                        SplineButton_Click                              'simulate click

                    END IF

                    IF MedianFilterBox.Checked THEN MedianFilter_Click
                    SegmentOCTimage                                     'perform auto Segmentation 

                END IF

                DrawSLOScan                                             'permanently draw the scan on the SLO image
                SuggestDiscMatchToReference                             'if reference set, then find best match to it and put crosshair on SLOimg

                cPos = sPos                                             'store current file position
                FormRepaint
                SetFocus(SaveButn.Handle)                               'wait for the save or skip button to be clicked
                SaveButn.Tag = False
                DO
                    DOEVENTS: SLEEP 0.05
                LOOP UNTIL (SaveButn.Tag = True)                        '"skip" will also trigger this

            WEND        'end of each series of images
        END IF      'end IF of myFileName <> ""
        

        myFileName = DIR$                                           'get next file in data folder, if none matching spec, returns null
    WEND            'end WHILE of myFileName <> ""



    IF OutputFileName <> "" THEN 
        IF MessageBox("Save data of " + Fname + " " + LName + CRLF + _
        "plotted on normative template?", "Save image...", &H30 + 4) = mrYes THEN NormForm_Save
    END IF


'    IF MessageBox("Completed analysis of all exported files in current folder. " + CRLF + _
'        "Delete Files in this folder and wait for another batch of export files?" + CRLF + _
'        "Clicking No will let you do more data analysis.", "Continue?", &H30 + 4) = mrYes THEN
'
'        IF MessageBox("Confirm: delete current xml and tif files?", "Cycle again?", &H30 + 4) = mrYes THEN
'            CHDIR MyDir
'            myFileName = DIR$("*.xml", 0)       'get HEYEX xml files
'            WHILE myFileName <> ""
'                KILL myFileName                 '-- KILL does not work with wildcards
'                DOEVENTS
'                myFileName = DIR$
'            WEND
'            myFileName = DIR$("*.tif", 0)       'get tif files
'            WHILE myFileName <> ""
'                KILL myFileName
'                DOEVENTS
'                myFileName = DIR$
'            WEND
'            ShowMessage "Working files removed. Load files into folder" + CRLF + _
'            CURDIR$ + CRLF + _
'            "then Click OK"
'        END IF
'
'    END IF

    IF MessageBox("Completed analysis of all exported files in current folder. " + CRLF + _
        "Continue with more data analysis?", "Continue?", &H30 + 4) = mrNo THEN Application.Terminate

LOOP UNTIL (GetXMLDataFolder = False)


Application.Terminate

